@@ -1,13 +1,17 @@
 import types
 import warnings
+from collections.abc import Sequence
 from functools import wraps as _wraps
 from typing import List
 
 import numpy as np
 import scipy.stats
-from scipy.special import softmax, xlogy
+import torch
+from scipy.special import xlogy
 from torch import Tensor
 
+from baal.utils.array_utils import to_prob
+
 available_reductions = {
     'max': lambda x: np.max(x, axis=tuple(range(1, x.ndim))),
     'min': lambda x: np.min(x, axis=tuple(range(1, x.ndim))),
@@ -59,28 +63,52 @@ def requireprobs(fn):
     @_wraps(fn)
     def wrapper(self, probabilities):
         # Expected shape : [n_sample, n_classes, ..., n_iterations]
-        bounded = np.min(probabilities) < 0 or np.max(probabilities) > 1.0
-        if bounded or not np.allclose(probabilities.sum(1), 1):
-            probabilities = softmax(probabilities, 1)
+        probabilities = to_prob(probabilities)
         return fn(self, probabilities)
 
     return wrapper
 
 
+def gather_expand(data, dim, index):
+    """
+    Gather indices `index` from `data` after expanding along dimension `dim`.
+
+    Args:
+        data (tensor): A tensor of data.
+        dim (int): dimension to expand along.
+        index (tensor): tensor with the indices to gather.
+
+    References:
+        Code from https://github.com/BlackHC/BatchBALD/blob/master/src/torch_utils.py
+
+    Returns:
+        Tensor with the same shape as `index`.
+    """
+    max_shape = [max(dr, ir) for dr, ir in zip(data.shape, index.shape)]
+    new_data_shape = list(max_shape)
+    new_data_shape[dim] = data.shape[dim]
+
+    new_index_shape = list(max_shape)
+    new_index_shape[dim] = index.shape[dim]
+
+    data = data.expand(new_data_shape)
+    index = index.expand(new_index_shape)
+
+    return torch.gather(data, dim, index)
+
+
 class AbstractHeuristic:
     """
     Abstract class that defines a Heuristic.
 
     Args:
         shuffle_prop (float): shuffle proportion.
-        threshold (Optional[float]): threshold the probabilities.
         reverse (bool): True if the most uncertain sample has the highest value.
         reduction (Union[str, Callable]): Reduction used after computing the score.
     """
 
-    def __init__(self, shuffle_prop=0.0, threshold=None, reverse=False, reduction='none'):
+    def __init__(self, shuffle_prop=0.0, reverse=False, reduction='none'):
         self.shuffle_prop = shuffle_prop
-        self.threshold = threshold
         self.reversed = reverse
         assert reduction in available_reductions or callable(reduction)
         self.reduction = reduction if callable(reduction) else available_reductions[reduction]
@@ -138,6 +166,26 @@ class AbstractHeuristic:
             scores[~np.isfinite(scores)] = fixed
         return scores
 
+    def reorder_indices(self, scores):
+        """
+        Order indices given their uncertainty score.
+
+        Args:
+            scores (ndarray/ List[ndarray]): Array of uncertainties or
+                list of arrays.
+
+        Returns:
+            ordered index according to the uncertainty (highest to lowes).
+        """
+        if isinstance(scores, Sequence):
+            scores = np.concatenate(scores)
+        assert scores.ndim <= 2
+        ranks = np.argsort(scores, -1)
+        if self.reversed:
+            ranks = ranks[::-1]
+        ranks = _shuffle_subset(ranks, self.shuffle_prop)
+        return ranks
+
     def get_ranks(self, predictions):
         """
         Rank the predictions according to their uncertainties.
@@ -154,14 +202,7 @@ class AbstractHeuristic:
         else:
             scores = self.get_uncertainties(predictions)
 
-        assert scores.ndim <= 2
-        ranks = np.argsort(scores, -1)
-        if self.threshold:
-            ranks = np.asarray([idx for idx in ranks if np.amax(predictions[idx]) > self.threshold])
-        if self.reversed:
-            ranks = ranks[::-1]
-        ranks = _shuffle_subset(ranks, self.shuffle_prop)
-        return ranks
+        return self.reorder_indices(scores)
 
     def __call__(self, predictions):
         """Rank the predictions according to their uncertainties."""
@@ -175,7 +216,6 @@ class BALD(AbstractHeuristic):
     Args:
         shuffle_prop (float): Amount of noise to put in the ranking. Helps with selection bias
             (default: 0.0).
-        threshold (Optional[Float]): Will ignore sample if the maximum prob is below this.
         reduction (Union[str, callable]): function that aggregates the results
             (default: 'none`).
 
@@ -183,9 +223,9 @@ class BALD(AbstractHeuristic):
         https://arxiv.org/abs/1703.02910
     """
 
-    def __init__(self, shuffle_prop=0.0, threshold=None, reduction='none'):
+    def __init__(self, shuffle_prop=0.0, reduction='none'):
         super().__init__(
-            shuffle_prop=shuffle_prop, threshold=threshold, reverse=True, reduction=reduction
+            shuffle_prop=shuffle_prop, reverse=True, reduction=reduction
         )
 
     @requireprobs
@@ -213,16 +253,20 @@ class BALD(AbstractHeuristic):
 
 class BatchBALD(BALD):
     """
-    Implementation of BatchBALD.
+    Implementation of BatchBALD from https://github.com/BlackHC/BatchBALD
 
     Args:
-        num_samples (int): Number of samples to select. (min 2*the amount of samples you want)
+        num_samples (int): Number of samples to select.
+        num_draw (int): Number of draw to perform from the history.
+                        From the paper `40000 // num_classes` is suggested.
         shuffle_prop (float): Amount of noise to put in the ranking. Helps with selection bias
             (default: 0.0).
-        threshold (Optional[Float]): Will ignore sample if the maximum prob is below this.
         reduction (Union[str, callable]): function that aggregates the results
             (default: 'none').
 
+    Notes:
+        This implementation only returns the ranking and not the score.
+
     References:
         https://arxiv.org/abs/1906.08158
 
@@ -231,27 +275,99 @@ class BatchBALD(BALD):
         Not tested on 4+ dims.
         """
 
-    def __init__(self, num_samples, shuffle_prop=0.0, threshold=None, reduction='none'):
+    def __init__(self, num_samples, num_draw=500, shuffle_prop=0.0, reduction='none'):
         self.epsilon = 1e-5
         self.num_samples = num_samples
-        super().__init__(shuffle_prop=shuffle_prop, threshold=threshold,
+        self.num_draw = num_draw
+        super().__init__(shuffle_prop=shuffle_prop,
                          reduction=reduction)
 
+    def _draw_choices(self, probs, n_choices):
+        """
+        Draw `n_choices` sample from `probs`.
+
+        References:
+            Code from https://github.com/BlackHC/BatchBALD/blob/master/src/torch_utils.py#L187
+
+        Returns:
+            choices: B... x `n_choices`
+
+        """
+        probs = probs.permute(0, 2, 1)
+        probs_B_C = probs.reshape((-1, probs.shape[-1]))
+
+        # samples: Ni... x draw_per_xx
+        choices = torch.multinomial(probs_B_C,
+                                    num_samples=n_choices, replacement=True)
+
+        choices_b_M = choices.reshape(list(probs.shape[:-1]) + [n_choices])
+        return choices_b_M.long()
+
+    def _sample_from_history(self, probs, num_draw=1000):
+        """
+        Sample `num_draw` choices from `probs`
+
+        Args:
+            probs (Tensor[batch, classes, ..., iterations]): Tensor to be sampled from.
+            num_draw (int): Number of draw.
+
+        References:
+            Code from https://github.com/BlackHC/BatchBALD/blob/master/src/joint_entropy/sampling.py
+
+        Returns:
+            Tensor[num_draw, iterations]
+        """
+        probs = torch.from_numpy(probs).double()
+
+        n_iterations = probs.shape[-1]
+
+        # [batch, draw, iterations]
+        choices = self._draw_choices(probs, num_draw)
+
+        # [batch, iterations, iterations, draw]
+        expanded_choices_N_K_K_S = choices[:, None, :, :]
+        expanded_probs_N_K_K_C = probs.permute(0, 2, 1)[:, :, None, :]
+
+        probs = gather_expand(expanded_probs_N_K_K_C, dim=-1, index=expanded_choices_N_K_K_S)
+        # exp sum log seems necessary to avoid 0s?
+        entropies = torch.exp(torch.sum(torch.log(probs), dim=0, keepdim=False))
+        entropies = entropies.reshape((n_iterations, -1))
+
+        samples_M_K = entropies.t()
+        return samples_M_K.numpy()
+
     def _conditional_entropy(self, probs):
         K = probs.shape[-1]
         return np.sum(-xlogy(probs, probs), axis=(1, -1)) / K
 
     def _joint_entropy(self, predictions, selected):
+        """
+        Compute the joint entropy between `preditions` and `selected`
+        Args:
+            predictions (Tensor): First tensor with shape [B, C, ..., Iterations]
+            selected (Tensor): Second tensor with shape [M, Iterations].
+
+        References:
+            Code from https://github.com/BlackHC/BatchBALD/blob/master/src/joint_entropy/sampling.py
+
+        Returns:
+            Generator yield B entropies.
+        """
         K = predictions.shape[-1]
+        C = predictions.shape[1]
+        B = predictions.shape[0]
         M = selected.shape[0]
+        predictions = predictions.swapaxes(1, 2)
 
         exp_y = np.array(
-            [np.matmul(selected, predictions[i].T) for i in range(predictions.shape[0])]) / K
+            [np.matmul(selected, predictions[i]) for i in range(predictions.shape[0])]) / K
+        assert exp_y.shape == (B, M, C)
         mean_entropy = selected.mean(-1, keepdims=True)[None]
+        assert mean_entropy.shape == (1, M, 1)
 
         step = 256
         for idx in range(0, exp_y.shape[0], step):
-            b_preds = exp_y[idx: (idx + step)]
+            b_preds = exp_y[idx:idx + step]
             yield np.sum(-xlogy(b_preds, b_preds) / mean_entropy, axis=(1, -1)) / M
 
     @requireprobs
@@ -265,29 +381,31 @@ class BatchBALD(BALD):
         Returns:
             Array of scores.
         """
-        MAX_SELECTED = 16000
         MIN_SPREAD = 0.1
         COUNT = 0
         # Get conditional_entropies_B
         conditional_entropies_B = self._conditional_entropy(predictions)
         bald_out = super().compute_score(predictions)
-        history = [self.reduction(bald_out).argmax()]
-        for step in range(2 * self.num_samples):
-            # Select M, iterations example from history, take entropy
-            selected = predictions[np.random.permutation(history)[:MAX_SELECTED]]
-            selected = self.reduction(scipy.stats.entropy(np.swapaxes(selected, 0, 1)))
+        # We start with the most uncertain sample according to BALD.
+        history = self.reduction(bald_out).argsort()[-1:].tolist()
+        for step in range(self.num_samples):
+            # Draw `num_draw` example from history, take entropy
+            # TODO use numpy/numba
+            selected = self._sample_from_history(predictions[history], num_draw=self.num_draw)
+
             # Compute join entropy
             joint_entropy = list(self._joint_entropy(predictions, selected))
             joint_entropy = np.concatenate(joint_entropy)
 
             partial_multi_bald_b = joint_entropy - conditional_entropies_B
             partial_multi_bald_b = self.reduction(partial_multi_bald_b)
-            partial_multi_bald_b[..., np.array(history)] = -1
+            partial_multi_bald_b[..., np.array(history)] = -1000
             # Add best to history
             partial_multi_bald_b = partial_multi_bald_b.squeeze()
             assert partial_multi_bald_b.ndim == 1
             winner_index = partial_multi_bald_b.argmax()
             history.append(winner_index)
+
             if partial_multi_bald_b.max() < MIN_SPREAD:
                 COUNT += 1
                 if COUNT > 50 or len(history) >= predictions.shape[0]:
@@ -295,6 +413,16 @@ class BatchBALD(BALD):
 
         return np.array(history)
 
+    def reorder_indices(self):
+        """This function is not supported by BatchBald.
+
+        Raises:
+            All the time.
+        """
+        raise Exception("BatchBald needs to have the whole pool at once,"
+                        "to be able to have relevant informationa chunk"
+                        " processing is not supported by BatchBald")
+
     def get_ranks(self, predictions):
         """
         Rank the predictions according to their uncertainties.
@@ -307,7 +435,6 @@ class BatchBALD(BALD):
 
         Raises:
             ValueError if predictions is a generator.
-
         """
         if isinstance(predictions, types.GeneratorType):
             raise ValueError("BatchBALD doesn't support generators.")
@@ -326,15 +453,14 @@ class Variance(AbstractHeuristic):
     Args:
         shuffle_prop (float): Amount of noise to put in the ranking. Helps with selection bias
             (default: 0.0).
-        threshold (Optional[Float]): Will ignore sample if the maximum prob is below this.
         reduction (Union[str, callable]): function that aggregates the results (default: `mean`).
     """
 
-    def __init__(self, shuffle_prop=0.0, threshold=None, reduction='mean'):
+    def __init__(self, shuffle_prop=0.0, reduction='mean'):
         _help = "Need to reduce the output from [n_sample, n_class] to [n_sample]"
         assert reduction != 'none', _help
         super().__init__(
-            shuffle_prop=shuffle_prop, threshold=threshold, reverse=True, reduction=reduction
+            shuffle_prop=shuffle_prop, reverse=True, reduction=reduction
         )
 
     def compute_score(self, predictions):
@@ -349,13 +475,12 @@ class Entropy(AbstractHeuristic):
     Args:
         shuffle_prop (float): Amount of noise to put in the ranking. Helps with selection bias
             (default: 0.0).
-        threshold (Optional[Float]): Will ignore sample if the maximum prob is below this.
         reduction (Union[str, callable]): function that aggregates the results (default: `none`).
     """
 
-    def __init__(self, shuffle_prop=0.0, threshold=None, reduction='none'):
+    def __init__(self, shuffle_prop=0.0, reduction='none'):
         super().__init__(
-            shuffle_prop=shuffle_prop, threshold=threshold, reverse=True, reduction=reduction
+            shuffle_prop=shuffle_prop, reverse=True, reduction=reduction
         )
 
     @singlepass
@@ -372,14 +497,13 @@ class Margin(AbstractHeuristic):
     Args:
         shuffle_prop (float): Amount of noise to put in the ranking. Helps with selection bias
             (default: 0.0).
-        threshold (Optional[Float]): Will ignore sample if the maximum prob is below this.
         reduction (Union[str, callable]): function that aggregates the results
             (default: `none`).
     """
 
-    def __init__(self, shuffle_prop=0.0, threshold=None, reduction='none'):
+    def __init__(self, shuffle_prop=0.0, reduction='none'):
         super().__init__(
-            shuffle_prop=shuffle_prop, threshold=threshold, reverse=False, reduction=reduction
+            shuffle_prop=shuffle_prop, reverse=False, reduction=reduction
         )
 
     @singlepass
@@ -395,13 +519,12 @@ class Certainty(AbstractHeuristic):
 
     Args:
         shuffle_prop (float): Amount of noise to put in the ranking. Helps with selection bias.
-        threshold (Optional[Float]): Will ignore sample if the maximum prob is below this.
         reduction (Union[str, callable]): function that aggregates the results.
     """
 
-    def __init__(self, shuffle_prop=0.0, threshold=None, reduction='none'):
+    def __init__(self, shuffle_prop=0.0, reduction='none'):
         super().__init__(
-            shuffle_prop=shuffle_prop, threshold=threshold, reverse=False, reduction=reduction
+            shuffle_prop=shuffle_prop, reverse=False, reduction=reduction
         )
 
     @singlepass
@@ -409,46 +532,58 @@ class Certainty(AbstractHeuristic):
         return np.max(predictions, axis=1)
 
 
-class Random(AbstractHeuristic):
-    """Random heuristic.
-
-    Args:
-        shuffle_prop (float): UNUSED
-        threshold (Optional[Float]): UNUSED
-        reduction (Union[str, callable]): UNUSED.
-    """
-
-    def __init__(self, shuffle_prop=0.0, threshold=None, reduction='none'):
-        super().__init__(1.0, threshold, False)
-
-    def get_ranks(self, predictions):
-        if isinstance(predictions, types.GeneratorType):
-            predictions = np.array([1 for _ in predictions])
-        ranks = np.arange(predictions.shape[0])
-        ranks = _shuffle_subset(ranks, self.shuffle_prop)
-        return ranks
-
-
 class Precomputed(AbstractHeuristic):
     """Precomputed heuristics.
 
     Args:
         shuffle_prop (float): Amount of noise to put in the ranking. Helps with selection bias.
-        threshold (Optional[Float]): Will ignore sample if the maximum prob is below this.
         reverse (Bool): Sort from lowest to highest if False.
     """
 
-    def __init__(self, shuffle_prop=0.0, threshold=None, reverse=False):
-        super().__init__(shuffle_prop, threshold, reverse=reverse)
+    def __init__(self, shuffle_prop=0.0, reverse=False):
+        super().__init__(shuffle_prop, reverse=reverse)
 
     def compute_score(self, predictions):
         return predictions
 
 
+class Random(Precomputed):
+    """Random heuristic.
+
+    Args:
+        shuffle_prop (float): UNUSED
+        reduction (Union[str, callable]): UNUSED.
+    """
+
+    def __init__(self, shuffle_prop=0.0, reduction='none'):
+        super().__init__(1.0, False)
+
+    def reorder_indices(self, predictions):
+        """
+        Order indices randomly.
+
+        Args:
+            predictions (ndarray): predictions for samples
+
+        Returns:
+            ranked indices (randomly)
+        """
+        if isinstance(predictions, Sequence):
+            predictions = np.concatenate(predictions)
+        ranks = np.arange(predictions.shape[0])
+        ranks = _shuffle_subset(ranks, self.shuffle_prop)
+        return ranks
+
+    def get_ranks(self, predictions):
+        if isinstance(predictions, types.GeneratorType):
+            predictions = np.array([1 for _ in predictions])
+        return self.reorder_indices(predictions)
+
+
 class CombineHeuristics(AbstractHeuristic):
     """Combine heuristics for multi-output models.
     heuristics would be applied on output predictions in the assigned order.
-    For each heuristic the necessary `reduction`, `reversed` and `threshold`
+    For each heuristic the necessary `reduction`, `reversed`
     parameters should be defined.
 
     NOTE: heuristics could be combined together only if they use the same
@@ -472,11 +607,11 @@ class CombineHeuristics(AbstractHeuristic):
        shuffle_prop (float): shuffle proportion.
 
     """
+
     def __init__(self, heuristics: List, weights: List, reduction='mean', shuffle_prop=0.0):
         super(CombineHeuristics, self).__init__(reduction=reduction, shuffle_prop=shuffle_prop)
         self.composed_heuristic = heuristics
         self.weights = weights
-
         reversed = [bool(heuristic.reversed) for heuristic in self.composed_heuristic]
 
         if all(item is False for item in reversed):
@@ -486,8 +621,6 @@ class CombineHeuristics(AbstractHeuristic):
         else:
             raise Exception("heuristics should have the same value for `revesed` parameter")
 
-        self.threshold = [bool(heuristic.threshold) for heuristic in self.composed_heuristic]
-
     def get_uncertainties(self, predictions):
         """
         Computes the score for each part of predictions according to the assigned heuristic.
@@ -512,20 +645,20 @@ class CombineHeuristics(AbstractHeuristic):
                 results.append(self.composed_heuristic[ind].get_uncertainties(prediction))
         return results
 
-    def get_ranks(self, predictions):
+    def reorder_indices(self, scores_list):
         """
-        Rank the predictions according to the weighted vote of each heuristic.
+        Order the indices based on the given scores.
 
         Args:
-            predictions (list[ndarray]):
-                list[[batch_size, C, ..., Iterations], [batch_size, C, ..., Iterations], ...]
+            scores_list (list(ndarray)/list(list(ndarray)):
 
         Returns:
-            Ranked index according to the uncertainty (highest to lowest).
+            ordered index according to the uncertainty (highest to lowes).
 
         """
-
-        scores_list = self.get_uncertainties(predictions)
+        if isinstance(scores_list[0], Sequence):
+            scores_list = list(zip(*scores_list))
+            scores_list = [np.concatenate(item) for item in scores_list]
 
         # normalizing weights
         w = np.array(self.weights).sum()
@@ -541,12 +674,24 @@ class CombineHeuristics(AbstractHeuristic):
         assert final_scores.ndim == 1
         ranks = np.argsort(final_scores)
 
-        for indx, threshold in enumerate(self.threshold):
-            if threshold:
-                ranks = np.asarray([idx for idx in ranks
-                                    if np.amax(predictions[indx][idx]) > threshold])
-
         if self.reversed:
             ranks = ranks[::-1]
         ranks = _shuffle_subset(ranks, self.shuffle_prop)
         return ranks
+
+    def get_ranks(self, predictions):
+        """
+        Rank the predictions according to the weighted vote of each heuristic.
+
+        Args:
+            predictions (list[ndarray]):
+                list[[batch_size, C, ..., Iterations], [batch_size, C, ..., Iterations], ...]
+
+        Returns:
+            Ranked index according to the uncertainty (highest to lowest).
+
+        """
+
+        scores_list = self.get_uncertainties(predictions)
+
+        return self.reorder_indices(scores_list)
